# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License,
# version 2.
package Plugins::RadiRuKo::FFLogger;

use strict;
use Carp::Assert;
use File::Temp;
use File::Spec;
use File::Monitor;
use Slim::Utils::Misc();
use Slim::Utils::Log;
use FileHandle;
use List::Util qw(first);
use POSIX qw(strftime);
use Time::HiRes q(time);
use Data::Dumper;

my $log   = logger('plugin.radiruko');
my $ffLog = Slim::Utils::Log->addLogCategory({
        category     => 'plugin.radiruko.ffmpeg',
        defaultLevel => 'WARN',
        description  => 'PLUGIN_RADIRUKO_FFMPEG',
    }
);

use constant INTERVAL => 1;

my $ffLogDir = eval { Slim::Utils::Misc::getTempDir() }    #8.2以降
    ||
    File::Temp::tempdir('ffLog-XXXXXX', CLEANUP => 1, DIR => File::Spec->tmpdir);

my $monitor = File::Monitor->new();

sub new ($) {
    my $class = shift;

    #program-YYYYMMDD-HHMMSS.log
    my $f = strftime('ffmpeg-%Y%m%d-%H%M%S.log', localtime());

    return bless {
        ffLogfile  => File::Spec->catfile($ffLogDir, $f),
        isWatching => 0,
    }, $class;
}

sub start ($) {
    my $self = shift;

    return if ($self->{isWatching});
    $self->{isWatching} = 1;

    $monitor->watch($self->{ffLogfile});
    $log->debug('watch: ' . $self->{ffLogfile});

    Slim::Utils::Timers::killTimers(__PACKAGE__, \&_do_watch);
    _do_watch();
}

sub stop ($) {
    my $self = shift;

    return unless ($self->{isWatching});
    $self->{isWatching} = 0;

    Slim::Utils::Timers::killTimers(__PACKAGE__, \&_do_watch);
    _do_watch();

    $monitor->unwatch($self->{ffLogfile});

    $log->debug('unwatch: ' . $self->{ffLogfile});

    unlink($self->{ffLogfile});
}

sub _ffreport_error_level () {
    return 48 if $ffLog->is_debug;
    return 32;
}

sub fflog_options ($) {
    return (qw(-nostats), -loglevel => 'level');
}

#
#環境変数FFREPORTの値
#
sub ffreport_environment_value ($) {
    my $self = shift;
    my $f    = $self->{ffLogfile};
    $f =~ s/([:\\])/\\$1/g;    #escape :\
    $f =~ s/%/%%/g;            #escape %

    my $lv = _ffreport_error_level();

    return "file=$f:level=$lv";
}

sub _parseStreamInfoLines {
    my $lines   = shift;
    my $streams = {
        Metadata => {},
    };

    my $stream_index;
    my $meta_indent;

    for (@$lines) {
        if (/^\s{2,4}(Stream (#0:\d+).* Audio: .+)$/) {
            $stream_index = $2;
            $meta_indent  = undef;
            local $_ = $1;
            my $audio_format  = $2 ? "$1/$2" : $1 if m{ Audio:\s+(\w+) *(?:\(([\w-]+)\))?};
            my ($sample_rate) = m{ (\d{4,}) Hz};
            my ($bitrate_kbs) = m{ (\d+) kb/s};

            $streams->{$stream_index} = {
                audio_format => $audio_format,
                sample_rate  => $sample_rate,
                bitrate      => $bitrate_kbs ? $bitrate_kbs * 1000 : undef,
                Metadata     => {},
            };
        } else {
            if (defined $meta_indent) {
                if (/^\s{$meta_indent}([\w\.]+)\s*:\s+(.+?)\s*$/) {
                    my $h = defined $stream_index ? $streams->{$stream_index} : $streams;
                    $h->{Metadata}{$1} = $2;
                } else {
                    $meta_indent = undef;
                }
            } else {
                #Input #0
                #  Metadata:
                # or
                #  Stream #0:0:
                #    Metadata:
                my $n = defined $stream_index ? '4,6' : '2';
                if (/^(\s{$n})Metadata:/) {
                    $meta_indent = length($1) + 2;
                }
            }
        }
    }

    return $streams;
}

sub _parseMappingLines {
    my $lines = shift;
    for (@$lines) {
        if (/Stream (#0:\d+) -> #0:0/) {
            return $1;
        }
    }
}

=pod
ストリーム情報を解析する
{
    'Selected' => '#0:0', #選択されたInput Stream
    'Input'    => {
        'Metadata' => { 'variant_bitrate' => '48676' },
        '#0:0'     => {
            'bitrate'      => 47000,
            'sample_rate'  => '48000',
            'audio_format' => 'aac (HE-AAC)',
            'Metadata'     => {
                'id3v2_priv.com.apple.streaming.transportStreamTimestamp' => '\\x00\\x00\\x00\\x01\\xea]\\x87\\x00',
                'variant_bitrate' => '48676'
            } }
    },
    'Output' => {
        'Metadata' => { 'encoder' => 'Lavf59.27.100' },
        '#0:0'     => {
            'bitrate'      => 47000,
            'sample_rate'  => '48000',
            'audio_format' => 'aac (HE-AAC)',
            'Metadata'     => {
                'id3v2_priv.com.apple.streaming.transportStreamTimestamp' =>
                    '\\x00\\x00\\x00\\x01\\xea]\\x87\\x00',
                'variant_bitrate' => '48676'
            } 
        }
    }
};
=cut

sub parseStreamInfo {
    my $self = shift;

    my %parsers = (
        Input    => [ qr{^Input #0,},       \&_parseStreamInfoLines ],
        Output   => [ qr{^Output #0,},      \&_parseStreamInfoLines ],
        Selected => [ qr{^Stream mapping:}, \&_parseMappingLines ],
    );

    my $fh = FileHandle->new($self->{ffLogfile}, '<:utf8') or do {
        logError("Couldn't open: $self->{ffLogfile}");
        return;
    };

    my %lines   = (_ => []);
    my $maxLine = 1024;
    while (my $line = <$fh> and $maxLine-- > 0) {
        #-loglevel levelのとき
        $line =~ s/^\[info\] //;

        if (my $k = first { $line =~ $parsers{$_}->[0] } keys %parsers) {
            $lines{_} = $lines{$k} = [];
        } else {
            push @{ $lines{_} }, $line;
        }
    }

    my $streamInfo = {};
    for my $k (keys %parsers) {
        my $func = $parsers{$k}->[1];
        $streamInfo->{$k} = $func->($lines{$k} || []);
    }

    $log->debug(Dumper $streamInfo);

    return $streamInfo;
}

sub _read_log ($) {
    my $fh = shift;
    my $lv = 'info';

    while (local $_ = <$fh>) {
        if (s/\[(debug|verbose|warning|info|error|fatal)\]\s//) {
            $lv = {
                warning => 'warn',
                verbose => 'debug',
            }->{$1} || $1;
        }
        $ffLog->$lv($_);
    }
}

sub _do_watch () {
    for my $d ($monitor->scan()) {
        next unless ($d->size > 0);

        $log->debug('fflog has been changed: ' . $d->name);

        my $fh = FileHandle->new($d->name, '<:utf8') or do {
            logError("Couldn't open: $d->name");
            next;
        };

        if ($d->is_created) {
            while (local $_ = <$fh>) {
                last if /^Command line:/;
            }
        } else {
            $fh->seek($d->old_size, 0);
        }

        _read_log($fh);
    }

    if ($monitor->has_monitors) {
        Slim::Utils::Timers::setTimer(__PACKAGE__, time() + INTERVAL, \&_do_watch);
    }
}

1
